@@ -2,6 +2,7 @@ package org.gcube.datacatalogue.grsf_manage_widget.client.view;
 
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Date;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Set;
@@ -22,8 +23,10 @@ import org.gcube.datacatalogue.grsf_manage_widget.client.view.subwidgets.Suggest
 import org.gcube.datacatalogue.grsf_manage_widget.shared.ConnectedBean;
 import org.gcube.datacatalogue.grsf_manage_widget.shared.HashTagsOnUpdate;
 import org.gcube.datacatalogue.grsf_manage_widget.shared.ManageProductBean;
+import org.gcube.datacatalogue.grsf_manage_widget.shared.RevertableOperationInfo;
 import org.gcube.datacatalogue.grsf_manage_widget.shared.SimilarGRSFRecord;
 import org.gcube.datacatalogue.grsf_manage_widget.shared.SourceRecord;
+import org.gcube.datacatalogue.grsf_manage_widget.shared.ex.GRSFRecordAlreadyManagedStatusException;
 import org.gcube.datacatalogue.grsf_manage_widget.shared.ex.NoGRSFRecordException;
 
 import com.github.gwtbootstrap.client.ui.AlertBlock;
@@ -39,6 +42,7 @@ import com.github.gwtbootstrap.client.ui.TextArea;
 import com.github.gwtbootstrap.client.ui.TextBox;
 import com.github.gwtbootstrap.client.ui.constants.AlertType;
 import com.github.gwtbootstrap.client.ui.constants.ButtonType;
+import com.github.gwtbootstrap.client.ui.constants.ControlGroupType;
 import com.github.gwtbootstrap.client.ui.event.HiddenEvent;
 import com.github.gwtbootstrap.client.ui.event.HiddenHandler;
 import com.google.gwt.core.client.GWT;
@@ -50,6 +54,7 @@ import com.google.gwt.event.dom.client.ChangeHandler;
 import com.google.gwt.event.dom.client.ClickEvent;
 import com.google.gwt.event.dom.client.ClickHandler;
 import com.google.gwt.event.shared.HandlerManager;
+import com.google.gwt.i18n.client.DateTimeFormat;
 import com.google.gwt.uibinder.client.UiBinder;
 import com.google.gwt.uibinder.client.UiField;
 import com.google.gwt.uibinder.client.UiHandler;
@@ -92,7 +97,7 @@ public class ManageProductWidget extends Composite{
 
 	@UiField
 	CheckBox traceabilityFlag;
-	
+
 	@UiField
 	CheckBox sdgFlag;
 
@@ -108,6 +113,12 @@ public class ManageProductWidget extends Composite{
 	@UiField
 	ControlGroup similarGRSFRecordGroup;
 
+	@UiField
+	ControlGroup suggestFurtherMerges;
+
+	@UiField
+	ControlGroup connectToOtherRecordsGroup;
+
 	@UiField
 	VerticalPanel panelForSimilarGRSFRecords;
 
@@ -147,9 +158,24 @@ public class ManageProductWidget extends Composite{
 	@UiField
 	Form formUpdate;
 
+	@UiField
+	VerticalPanel moreInfoAboutOperation;
+
 	@UiField
 	Image loadingImage;
 
+	@UiField
+	TextArea requestAuthor;
+
+	//	@UiField
+	//	TextArea requestTypeBox;
+
+	@UiField
+	TextArea requestRecordUUID;
+
+	@UiField
+	TextArea requestTimestamp;
+
 	public static final String LOADING_IMAGE_URL = GWT.getModuleBaseURL() + "../images/loader.gif";
 
 	// messages used here and there
@@ -165,29 +191,32 @@ public class ManageProductWidget extends Composite{
 
 	// the objects to be managed
 	private ManageProductBean bean;
+	private RevertableOperationInfo revertableOperation = null;
 	private final static List<Status> STATUS = new ArrayList<Status>(Arrays.asList(Status.values()));
 
 	// similar records and to connect widgets references
 	private SimilarGRSFRecordWidget similarRecordPanel;
 	private SuggestMerges suggestedMergesPanel;
 	private ConnectToWidget connectWidget;
-
 	private boolean updateSucceeded = false;
+	private boolean isRevertingMerge = false;
 
 	/**
-	 * Build a ManageProduct widget for the product with the specified id.
-	 * @param productIdentifier
+	 * Build a ManageProduct widget for the product with the specified id or with a given revert merge url.
+	 * @param productIdentifierOrUrl
 	 * @param eventBus
 	 */
-	public ManageProductWidget(String productIdentifier, HandlerManager eventBus) {
+	public ManageProductWidget(String productIdentifierOrUrl, HandlerManager eventBus) {
 		initWidget(uiBinder.createAndBindUi(this));
 		this.eventBus = eventBus;
 
-		GWT.log("item identifier is " + productIdentifier);
-
-		if(productIdentifier == null || productIdentifier.isEmpty())
+		if(productIdentifierOrUrl == null || productIdentifierOrUrl.isEmpty())
 			return;
 
+		// check if we are going to perform a revert merge operation
+		if(productIdentifierOrUrl.startsWith("http"))
+			isRevertingMerge = true;
+
 		// start loader service 
 		loadingImage.setUrl(LOADING_IMAGE_URL);
 		loadingImage.setVisible(true);
@@ -196,6 +225,7 @@ public class ManageProductWidget extends Composite{
 
 		// show modal
 		manageProductModal.addStyleName("management-metadata-modal-style");
+		manageProductModal.addStyleName("modal-top-custom");
 		((Element)manageProductModal.getElement().getChildNodes().getItem(1)).addClassName("modal-body-custom");
 		manageProductModal.show();
 
@@ -210,7 +240,10 @@ public class ManageProductWidget extends Composite{
 		});
 
 		// async request to fetch the product
-		retrieveProductBean(productIdentifier);
+		if(!isRevertingMerge)
+			retrieveProductBean(productIdentifierOrUrl);
+		else
+			validateRevertUrlBefore(productIdentifierOrUrl);
 		listenEvents(this.eventBus);
 	}
 
@@ -229,12 +262,55 @@ public class ManageProductWidget extends Composite{
 		});
 	}
 
+	/**
+	 * Validate the parameters of the url and ask the editor/reviewer what he/she wants to do.
+	 * @param encryptedUrlOperation
+	 */
+	private void validateRevertUrlBefore(String encryptedUrlOperation) {
+
+		confirmButton.setEnabled(false);
+
+		service.validateRevertOperation(encryptedUrlOperation, new AsyncCallback<RevertableOperationInfo>() {
+
+			@Override
+			public void onSuccess(RevertableOperationInfo result) {
+
+				loadingImage.setVisible(false);
+
+				if(result != null){
+					revertableOperation = result;
+					String dateString = DateTimeFormat.getFormat("MM/dd/yyyy HH:mm:ss (z)").format(new Date(revertableOperation.getTimestamp()));					
+					requestAuthor.setText(revertableOperation.getFullNameOriginalAdmin() + "(" + revertableOperation.getUserNameOriginalAdmin() + ")");
+					//					requestTypeBox.setText(revertableOperation.getOperation().toString().toUpperCase());
+					requestRecordUUID.setText(revertableOperation.getUuid());
+					requestTimestamp.setText(dateString);
+					moreInfoAboutOperation.setVisible(true);
+
+					// proceed by fetching the other details
+					retrieveProductBean(revertableOperation.getUuid());
+
+				}else{
+					showInfo("Unable to evaluate the url for this operation. ", AlertType.ERROR);
+					formUpdate.setVisible(false);
+					loadingImage.setVisible(false);
+				}
+			}
+
+			@Override
+			public void onFailure(Throwable caught) {
+				showInfo("Unable to perform this operation. " + (caught != null ? "Error was " + caught.getMessage() : ""), AlertType.ERROR);
+				formUpdate.setVisible(false);
+				loadingImage.setVisible(false);
+			}
+		});
+
+	}
 
 	/**
 	 * Actually builds the widget... asks for details about the record
-	 * @param productIdentifier
+	 * @param productIdentifierOrUrl
 	 */
-	private void retrieveProductBean(final String productIdentifier) {
+	private void retrieveProductBean(final String productIdentifierOrUrl) {
 
 		// check if it is an administrator
 		service.isAdminUser(new AsyncCallback<Boolean>() {
@@ -257,7 +333,7 @@ public class ManageProductWidget extends Composite{
 						GWT.log("Hide management panel event sent");
 					}
 				}else{
-					service.getProductBeanById(productIdentifier, new AsyncCallback<ManageProductBean>() {
+					service.getProductBeanById(productIdentifierOrUrl, isRevertingMerge, new AsyncCallback<ManageProductBean>() {
 
 						@Override
 						public void onSuccess(ManageProductBean resBean) {
@@ -282,17 +358,17 @@ public class ManageProductWidget extends Composite{
 								});
 
 								semanticIdentifierTextBox.setText(bean.getSemanticIdentifier());
-								
+
 								// update product type listbox.. get the available types for the record under management
 								List<String> types = getTypesForRecord(bean.getDomain());
 								productGrsfTypeListbox.addItem(bean.getCurrentGrsfType(), bean.getCurrentGrsfType());
-								
+								types.remove(bean.getCurrentGrsfType());
 								for (String type : types) {
 									productGrsfTypeListbox.addItem(type, type);
 								}
-								
+
 								//select the current
- 								productGrsfTypeListbox.setSelectedValue(bean.getCurrentGrsfType());
+								productGrsfTypeListbox.setSelectedValue(bean.getCurrentGrsfType());
 
 								productGrsfTypeListbox.addChangeHandler(new ChangeHandler() {
 
@@ -302,7 +378,7 @@ public class ManageProductWidget extends Composite{
 									}
 								});
 								currentStatus.setText(bean.getCurrentStatus().toString());
-								
+
 								// traceability flag
 								traceabilityFlag.setValue(bean.isTraceabilityFlag());
 								traceabilityFlag.setText("Traceability");
@@ -314,7 +390,7 @@ public class ManageProductWidget extends Composite{
 										eventBus.fireEvent(new EnableConfirmButtonEvent());
 									}
 								});
-								
+
 								// sdg flag
 								sdgFlag.setValue(bean.isSdgFlag());
 								sdgFlag.setText("Sustainable Development Goals");
@@ -326,7 +402,7 @@ public class ManageProductWidget extends Composite{
 										eventBus.fireEvent(new EnableConfirmButtonEvent());
 									}
 								});
-								
+
 								// manage sources
 								List<SourceRecord> availableSources = bean.getSources();
 								panelForSourceItems.add(new SourceWidget(availableSources));
@@ -350,14 +426,36 @@ public class ManageProductWidget extends Composite{
 								// check for new status box
 								List<Status> statusToShow = new ArrayList<Status>(STATUS);
 								statusToShow.remove(bean.getCurrentStatus());
-								
-								// remove to be merged, since it cannot be set by a user
-								statusToShow.remove(Status.To_be_Merged); 
 
-								// if the record isn't approved, then remove also archived
-								if(!bean.getCurrentStatus().equals(Status.Approved))
-									statusToShow.remove(Status.Archived);
+								// if we are going to revert an operation
+								if(isRevertingMerge){
+									statusToShow.clear();
+									statusToShow.add(Status.Reject_Merge);
 
+									// all other stuff must be frozen, since the only allowed operation is to reject the merge
+									shortNameTextBox.setEnabled(false);
+									productGrsfTypeListbox.setEnabled(false);
+									sdgFlag.setEnabled(false);
+									traceabilityFlag.setEnabled(false);
+
+									// freeze other panels
+									connectWidget.freezeWidget();
+									similarRecordPanel.freezeWidget();
+									suggestedMergesPanel.freezeWidget();
+
+								}else{
+
+									// remove to be merged, since it cannot be set by a user
+									statusToShow.remove(Status.To_be_Merged); 
+
+									// remove reject merge
+									statusToShow.remove(Status.Reject_Merge);
+
+									// if the record isn't approved, then remove also archived
+									if(!bean.getCurrentStatus().equals(Status.Approved))
+										statusToShow.remove(Status.Archived);
+
+								}
 								listBoxStatus.addItem("Select a new status");
 								listBoxStatus.getElement().<SelectElement>cast().getOptions().getItem(0).setDisabled(true);
 								for (Status availableStatus : statusToShow) {
@@ -392,8 +490,10 @@ public class ManageProductWidget extends Composite{
 
 							if(caught instanceof NoGRSFRecordException)
 								showInfo(NO_GRSF_RECORD_BEAN, AlertType.WARNING);
-							else
-								showInfo("Error is " + caught, AlertType.ERROR);
+							else if(caught instanceof GRSFRecordAlreadyManagedStatusException) {
+								showInfo("WARNING: "+caught.getMessage(), AlertType.WARNING);
+							}else
+								showInfo("Error is " + caught.getMessage(), AlertType.ERROR);
 
 							// hide the form and disable the send button
 							formUpdate.setVisible(false);
@@ -423,22 +523,104 @@ public class ManageProductWidget extends Composite{
 	@UiHandler("confirmButton")
 	void onSaveButton(ClickEvent ce){
 
-		String report = "";
-		Set<String> hashtags = new HashSet<>();
+		annotationAreaGroup.setType(ControlGroupType.NONE);
+		annotationArea.setPlaceholder("");
+		listBoxStatusGroup.setType(ControlGroupType.NONE);
+		
+		if(isRevertingMerge && listBoxStatus.getSelectedIndex() <= 0){
+			listBoxStatusGroup.setType(ControlGroupType.ERROR);
+			listBoxStatus.setFocus(true);
+			return;
+		}
+
+		// in case we are rejecting the record or the merge, the annotation is mandatory
+		Status selectedStatus = Status.fromString(listBoxStatus.getSelectedItemText());
+		if(selectedStatus.equals(Status.Reject_Merge) || selectedStatus.equals(Status.Rejected)){
+			if(annotationArea.getText() == null || annotationArea.getText().isEmpty()){
+				annotationArea.setPlaceholder("In case of reject operations you must specify an annotation message.");
+				annotationAreaGroup.setType(ControlGroupType.ERROR);
+				annotationArea.setFocus(true);
+				return;
+			}
+		}
 
 		manageProductModal.setCloseVisible(false);
 		cancelButton.setEnabled(false);
 		confirmButton.setEnabled(false);
 		loaderIcon.setVisible(true);
 
+		if(isRevertingMerge)
+			revertMergeOperation();
+		else
+			performRecordUpdate();
+	}
+
+	/**
+	 * Perform a revert merge operation
+	 */
+	private void revertMergeOperation(){
+		service.performRevertOperation(revertableOperation, new AsyncCallback<Boolean>() {
+
+			@Override
+			public void onSuccess(Boolean result) {
+
+				confirmButton.setEnabled(true);
+				loaderIcon.setVisible(false);
+
+				if(!result){
+					showInfo("Unable to perform this operation. ", AlertType.ERROR);
+				}
+				else{
+					infoBlock.setVisible(true);
+					infoBlock.setType(AlertType.SUCCESS);
+					infoBlock.setText("The request has been processed successfully!");
+					confirmButton.removeFromParent();
+					cancelButton.setText("Ok");
+					cancelButton.setType(ButtonType.INFO);
+					cancelButton.setEnabled(true);
+				}
+			}
+
+			@Override
+			public void onFailure(Throwable caught) {
+				showInfo("Unable to perform this operation. " + (caught != null ? "Error was " + caught : ""), AlertType.ERROR);
+				confirmButton.setEnabled(true);
+				loadingImage.setVisible(false);
+			}
+		});
+
+	}
+
+
+	/**
+	 * Perform an update for this record
+	 */
+	private void performRecordUpdate(){
+		String report = "";
+		Set<String> hashtags = new HashSet<>();
+
 		// get short name
 		bean.setShortNameUpdated(shortNameTextBox.getText());
-
+		report += "\n-Information of the record managed:";
+		report += "\n\t- GRSF Name '" + bean.getTitle() + "' ;";
+		report += "\n\t- Short Name '" + bean.getShortName() + "' ;";
+		report += "\n\t- URL '" + bean.getUrl() + "' ;";
+		report += "\n\t- Semantic Identifier '" + bean.getSemanticIdentifier() + "' ;";
+		
+		String involvedSourceRecords = "\n\t- Database Sources involved: ";
+		List<SourceRecord> sources = bean.getSources();
+		for (SourceRecord sourceRecord : sources) {
+			involvedSourceRecords += sourceRecord.getName() + " ";
+		}
+		
+		involvedSourceRecords += ";";
+		report += involvedSourceRecords;
+		
 		if(!bean.getShortName().equals(bean.getShortNameUpdated())){
 			report += "\n- The GRSF Short Name has been changed to '" + bean.getShortNameUpdated() + "' from '" + bean.getShortName() + "';";
 			hashtags.add(HashTagsOnUpdate.SHORTNAME_UPDATED.getString());
 		}
-		
+
 		// status
 		bean.setNewGrsfType(productGrsfTypeListbox.getSelectedItemText());
 		if(bean.getNewGrsfType().equalsIgnoreCase(bean.getCurrentGrsfType())){
@@ -456,16 +638,22 @@ public class ManageProductWidget extends Composite{
 		if(!bean.getConnections().isEmpty()){
 			report += "\n- Suggested connections:";
 			boolean addConnectionHashtag = false;
+			boolean removeConnectionHashtag = false;
 			for(ConnectedBean cb: bean.getConnections()){
-				if(cb.isRemove())
-					report += "\n\t - remove connection with record " + cb.getKnowledgeBaseId() + ";";
+				if(cb.isRemove()){
+					removeConnectionHashtag = true;
+					report += "\n\t - remove connection with record " + cb.getUrl() + ";";
+				}
 				else if(cb.isConnect()){
 					addConnectionHashtag = true;
-					report += "\n\t - add connection with record " + cb.getKnowledgeBaseId() + ";";
+					report += "\n\t - add connection with record " + cb.getUrl() + ";";
 				}else
-					report += "\n\t - keep this suggestion " + cb.getKnowledgeBaseId() + ";";
+					report += "\n\t - keep this suggestion " + cb.getUrl() + ";";
 			}
 			
+			if(removeConnectionHashtag)
+				hashtags.add(HashTagsOnUpdate.DISCONNECT.getString());
+				
 			if(addConnectionHashtag)
 				hashtags.add(HashTagsOnUpdate.CONNECT.getString());
 		}
@@ -485,7 +673,11 @@ public class ManageProductWidget extends Composite{
 			for(SimilarGRSFRecord sR: bean.getSimilarGrsfRecords()){
 				if(sR.isSuggestedMerge()){
 					bean.setMergesInvolved(true);
-					report += "\n\t - merge the current record with record " + sR.getKnowledgeBaseId() + ";";
+					report += "\n\t - merge the current record with record '" + sR.getTitle() + "' ;";
+					report += "\n\t \t- GRSF Name '" + sR.getTitle() + "' ;";
+					report += "\n\t \t- Short Name '" + sR.getShortName() + "' ;";
+					report += "\n\t \t- URL '" + sR.getUrl() + "' ;";
+					report += "\n\t \t- Semantic Identifier '" + sR.getSemanticIdentifier() + "' ;";
 				}
 			}
 			if(bean.isMergesInvolved()){
@@ -493,7 +685,7 @@ public class ManageProductWidget extends Composite{
 				hashtags.add(HashTagsOnUpdate.MERGE.getString());
 			}
 		}
-		
+
 		// set new values
 		bean.setAnnotation(new HTML(annotationArea.getText().trim()).getText());
 
@@ -510,10 +702,10 @@ public class ManageProductWidget extends Composite{
 			else
 				hashtags.add(HashTagsOnUpdate.TRACEABILITY_FLAG_UNSET.getString());
 		}
-		
+
 		// update the traceability flag
 		bean.setTraceabilityFlag(traceabilityNewValue);
-		
+
 		// sdg flag
 		Boolean sdgNewValue = sdgFlag.getValue();
 		boolean currentSdgFlag = bean.isSdgFlag();
@@ -524,10 +716,10 @@ public class ManageProductWidget extends Composite{
 			else
 				hashtags.add(HashTagsOnUpdate.SDG_FLAG_UNSET.getString());
 		}
-		
+
 		// update the traceability flag
 		bean.setSdgFlag(sdgNewValue);
-		
+
 		// force the new status in the listbox
 		if(bean.isMergesInvolved()){
 			bean.setNewStatus(Status.To_be_Merged);
@@ -577,7 +769,6 @@ public class ManageProductWidget extends Composite{
 				showInfo(STATUS_UPDATE_ERROR + ": " + caught.getMessage(), AlertType.ERROR);
 			}
 		});
-
 	}
 
 	@UiHandler("cancelButton")
@@ -596,8 +787,9 @@ public class ManageProductWidget extends Composite{
 		infoBlock.setText(statusUpdateError);
 		infoBlock.setType(type);
 		infoBlock.setVisible(true);
+		infoBlock.getElement().focus();
 	}
-	
+
 	/**
 	 * Retrieve the list of types for stocks and fisheries, given the domain
 	 * @param domain
